/* $XFree86: xc/programs/Xserver/hw/xfree86/int10/helper_exec.c,v 1.16 2001/04/30 14:34:57 tsi Exp $ */
/*
 *                   XFree86 int10 module
 *   execute BIOS int 10h calls in x86 real mode environment
 *                 Copyright 1999 Egbert Eich
 *
 *   Part of this is based on code taken from DOSEMU
 *   (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team"
 */

/*
 * To debug port accesses define PRINT_PORT.
 * Note! You also have to comment out ioperm()
 * in xf86EnableIO(). Otherwise we won't trap
 * on PIO.
 */
#include <x86emu.h>
#include "post.h"
#include <stdio.h>
#include <asm/io.h>
#include <sys/time.h>

extern void *iobase;

//#define IOTRACE(fmt...) fprintf(stderr, fmt)
//#define IOTRACE(fmt...) printf(fmt)
#define IOTRACE(fmt...)
#define __iomem

static unsigned int inb40time;

#ifdef __powerpc__
static inline int in_8(volatile unsigned char __iomem *addr)
{
	int ret;

	__asm__ __volatile__(
		"lbz%U1%X1 %0,%1;\n"
		"twi 0,%0,0;\n"
		"isync" : "=r" (ret) : "m" (*addr));
	return ret;
}

static inline void out_8(volatile unsigned char __iomem *addr, int val)
{
	__asm__ __volatile__("stb%U0%X0 %1,%0; eieio" : "=m" (*addr) : "r" (val));
}

static inline int in_le16(volatile unsigned short __iomem *addr)
{
	int ret;

	__asm__ __volatile__("lhbrx %0,0,%1;\n"
			     "twi 0,%0,0;\n"
			     "isync" : "=r" (ret) :
			      "r" (addr), "m" (*addr));
	return ret;
}

static inline int in_be16(volatile unsigned short __iomem *addr)
{
	int ret;

	__asm__ __volatile__("lhz%U1%X1 %0,%1;\n"
			     "twi 0,%0,0;\n"
			     "isync" : "=r" (ret) : "m" (*addr));
	return ret;
}

static inline void out_le16(volatile unsigned short __iomem *addr, int val)
{
	__asm__ __volatile__("sthbrx %1,0,%2; eieio" : "=m" (*addr) :
			      "r" (val), "r" (addr));
}

static inline void out_be16(volatile unsigned short __iomem *addr, int val)
{
	__asm__ __volatile__("sth%U0%X0 %1,%0; eieio" : "=m" (*addr) : "r" (val));
}

static inline unsigned in_le32(volatile unsigned __iomem *addr)
{
	unsigned ret;

	__asm__ __volatile__("lwbrx %0,0,%1;\n"
			     "twi 0,%0,0;\n"
			     "isync" : "=r" (ret) :
			     "r" (addr), "m" (*addr));
	return ret;
}

static inline unsigned in_be32(volatile unsigned __iomem *addr)
{
	unsigned ret;

	__asm__ __volatile__("lwz%U1%X1 %0,%1;\n"
			     "twi 0,%0,0;\n"
			     "isync" : "=r" (ret) : "m" (*addr));
	return ret;
}

static inline void out_le32(volatile unsigned __iomem *addr, int val)
{
	__asm__ __volatile__("stwbrx %1,0,%2; eieio" : "=m" (*addr) :
			     "r" (val), "r" (addr));
}

static inline void out_be32(volatile unsigned __iomem *addr, int val)
{
	__asm__ __volatile__("stw%U0%X0 %1,%0; eieio" : "=m" (*addr) : "r" (val));
}
#endif

extern void *vmem;

/* compute a pointer. This replaces code scattered all over the place! */
static u8 *x_mem_ptr(u32 addr, int size)
{
	u8 *retaddr = 0;

	if (addr >= 0xa0000 && addr < 0xb8000) {
#if 0
		int shift = addr & 1;
		addr &= ~1;
		return (addr - 0xa0000) * 4 + shift + 2 + vmem;
#else
		return (addr - 0xa0000) * 4 + 2 + vmem;
#endif
	}
	if (addr >= 0xb8000 && addr < 0xc0000) {
		int shift = addr & 1;
		addr &= 0xfffffffe;
		return (addr - 0xb8000) * 4 + shift + vmem;
	}

	if (addr > M.mem_size - size) {
		printf("mem_ptr: address %lx out of range!\n",
		       (unsigned long)addr);
		HALT_SYS();
	}
	if (addr < 0x200) {
		//printk("%x:%x updating int vector 0x%x\n",
		//       M.x86.R_CS, M.x86.R_IP, addr >> 2);
	}
	retaddr = (u8 *) (M.mem_base + addr);

	return retaddr;
}

/****************************************************************************
PARAMETERS:
addr	- Emulator memory address to read

RETURNS:
Byte value read from emulator memory.

REMARKS:
Reads a byte value from the emulator memory. 
****************************************************************************/
u8 X86API x_rdb(u32 addr)
{
	u8 val;
	u8 *ptr;

	ptr = x_mem_ptr(addr, 1);

#if __powerpc__
	val = in_8(ptr);
#else
	val = *ptr;
#endif
	return val;
}

/****************************************************************************
PARAMETERS:
addr	- Emulator memory address to read

RETURNS:
Word value read from emulator memory.

REMARKS:
Reads a word value from the emulator memory.
****************************************************************************/
u16 X86API x_rdw(u32 addr)
{
	u16 val = 0;
	u16 *ptr;

	ptr = (u16 *)x_mem_ptr(addr, 2);
#if __powerpc__
	val = in_le16(ptr);
#else
	val = *ptr;
#endif
	return val;
}

/****************************************************************************
PARAMETERS:
addr	- Emulator memory address to read

RETURNS:
Long value read from emulator memory.
REMARKS:
Reads a long value from the emulator memory. 
****************************************************************************/
u32 X86API x_rdl(u32 addr)
{
	u32 val = 0;
	u32 *ptr;

	ptr = (u32 *)x_mem_ptr(addr, 4);
#if __powerpc__
	val = in_le32(ptr);
#else
       	val = *ptr;
#endif
	return val;
}

/****************************************************************************
PARAMETERS:
addr	- Emulator memory address to read
val		- Value to store

REMARKS:
Writes a byte value to emulator memory.
****************************************************************************/
void X86API x_wrb(u32 addr, u8 val)
{
	u8 *ptr;

	ptr = x_mem_ptr(addr, 1);
#ifdef __powerpc__
	out_8(ptr, val);
#else
	*(u8 *) (ptr) = val;
#endif
}

/****************************************************************************
PARAMETERS:
addr	- Emulator memory address to read
val		- Value to store

REMARKS:
Writes a word value to emulator memory.
****************************************************************************/
void X86API x_wrw(u32 addr, u16 val)
{
	u16 *ptr;

	ptr = (u16 *)x_mem_ptr(addr, 2);
#if __powerpc__
	out_le16(ptr, val);
#else
	*ptr = val;
#endif
}

/****************************************************************************
PARAMETERS:
addr	- Emulator memory address to read
val		- Value to store

REMARKS:
Writes a long value to emulator memory. 
****************************************************************************/
void X86API x_wrl(u32 addr, u32 val)
{
	u32 *ptr;

	ptr = (u32 *)x_mem_ptr(addr, 4);
#if __powerpc__
	out_le32(ptr, val);
#else
	*(u32 *) (ptr) = val;
#endif
}

/* general software interrupt handler */
u32 getIntVect(int num)
{
	return MEM_RW(num << 2) + (MEM_RW((num << 2) + 2) << 4);
}

void pushw(u16 val)
{
	X86_ESP -= 2;
	MEM_WW(((u32) X86_SS << 4) + X86_SP, val);
}

int run_bios_int(int num)
{
	u32 eflags;

	eflags = X86_EFLAGS;
	pushw(eflags);
	pushw(X86_CS);
	pushw(X86_IP);
	X86_CS = MEM_RW((num << 2) + 2);
	X86_IP = MEM_RW(num << 2);

	printf("%s: INT %x CS:IP = %x:%x\n", __FUNCTION__,
	       num, MEM_RW((num << 2) + 2), MEM_RW(num << 2));

	return 1;
}

int port_rep_inb(u16 port, u32 base, int d_f, u32 count)
{
	register int inc = d_f ? -1 : 1;
	u32 dst = base;
	while (count--) {
		MEM_WB(dst, x_inb(port));
		dst += inc;
	}
	return dst - base;
}

int port_rep_inw(u16 port, u32 base, int d_f, u32 count)
{
	register int inc = d_f ? -2 : 2;
	u32 dst = base;
	while (count--) {
		MEM_WW(dst, x_inw(port));
		dst += inc;
	}
	return dst - base;
}

int port_rep_inl(u16 port, u32 base, int d_f, u32 count)
{
	register int inc = d_f ? -4 : 4;
	u32 dst = base;
	while (count--) {
		MEM_WL(dst, x_inl(port));
		dst += inc;
	}
	return dst - base;
}

int port_rep_outb(u16 port, u32 base, int d_f, u32 count)
{
	register int inc = d_f ? -1 : 1;
	u32 dst = base;
	while (count--) {
		x_outb(port, MEM_RB(dst));
		dst += inc;
	}
	return dst - base;
}

int port_rep_outw(u16 port, u32 base, int d_f, u32 count)
{
	register int inc = d_f ? -2 : 2;
	u32 dst = base;
	while (count--) {
		x_outw(port, MEM_RW(dst));
		dst += inc;
	}
	return dst - base;
}

int port_rep_outl(u16 port, u32 base, int d_f, u32 count)
{
	register int inc = d_f ? -4 : 4;
	u32 dst = base;
	while (count--) {
		x_outl(port, MEM_RL(dst));
		dst += inc;
	}
	return dst - base;
}

u8 x_inb(u16 port)
{
	u8 val;

	IOTRACE("inb(0x%04x) = ", port);

	if (port == 0x40) {
		inb40time++;
		return inb40time >> ((inb40time & 1) << 3);
	}
	if (port < 0x100)
		return 0xff;

#ifdef __powerpc__
	val = in_8(iobase + port);
#else
	val = inb(port);
#endif
	IOTRACE("0x%02x\n", val);

	return val;
}

u16 x_inw(u16 port)
{
	u16 val;

	IOTRACE("inw(0x%04x) = ", port);

	if (port == 0x5c) {
		struct timeval tv;
		gettimeofday(&tv, NULL);
		return tv.tv_usec / 3;
	}
	if (port < 0x100)
		return 0xffff;

#ifdef __powerpc__
	val = in_le16(iobase + port);
#else
	val = inw(port);
#endif

	IOTRACE("0x%04x\n", val);
	return val;
}

u32 x_inl(u16 port)
{
	u32 val;

	IOTRACE("inl(0x%04x) = ", port);

	if (port < 0x100)
		return 0xffffffff;

#ifdef __powerpc__
	val = in_le32(iobase + port);
#else
	val = inl(port);
#endif

	IOTRACE("0x%08x\n", val);
	return val;
}

void x_outb(u16 port, u8 val)
{
	IOTRACE("outb(0x%02x, 0x%04x)\n", val, port);

    if ((port == 0x43) && (val == 0)) {
	    struct timeval tv;
	    gettimeofday(&tv, NULL);
	    inb40time = tv.tv_usec | 1;
    }
    if (port < 0x100)
	    return;

#ifdef __powerpc__
	out_8(iobase + port, val);
#else
	outb(val, port);
#endif
}

void x_outw(u16 port, u16 val)
{
	IOTRACE("outw(0x%04x, 0x%04x)\n", val, port);

    if (port < 0x100)
	    return;

#ifdef __powerpc__
	out_le16(iobase + port, val);
#else
	outw(val, port);
#endif
}

void x_outl(u16 port, u32 val)
{
	IOTRACE("outl(0x%08x, 0x%04x)\n", val, port);

    if (port < 0x100)
	    return;

#ifdef __powerpc__
	out_le32(iobase + port, val);
#else
	outl(val, port);

#endif
}

#if 0
u8 Mem_rb(int addr)
{
	return (*current->mem->rb) (current, addr);
}

u16 Mem_rw(int addr)
{
	return (*current->mem->rw) (current, addr);
}

u32 Mem_rl(int addr)
{
	return (*current->mem->rl) (current, addr);
}

void Mem_wb(int addr, u8 val)
{
	(*current->mem->wb) (current, addr, val);
}

void Mem_ww(int addr, u16 val)
{
	(*current->mem->ww) (current, addr, val);
}

void Mem_wl(int addr, u32 val)
{
	(*current->mem->wl) (current, addr, val);
}
#endif

void getsecs(unsigned long *sec, unsigned long *usec)
{
	struct timeval tv;
	gettimeofday(&tv, 0);
	*sec = tv.tv_sec;
	*usec = tv.tv_usec;
}

#define TAG(Cfg1Addr) (Cfg1Addr & 0xffff00)
#define OFFSET(Cfg1Addr) (Cfg1Addr & 0xff)

u8 bios_checksum(u8 * start, int size)
{
	u8 sum = 0;

	while (size-- > 0)
		sum += *start++;
	return sum;
}

/*
 * Lock/Unlock legacy VGA. Some Bioses try to be very clever and make
 * an attempt to detect a legacy ISA card. If they find one they might
 * act very strange: for example they might configure the card as a
 * monochrome card. This might cause some drivers to choke.
 * To avoid this we attempt legacy VGA by writing to all know VGA
 * disable registers before we call the BIOS initialization and
 * restore the original values afterwards. In beween we hold our
 * breath. To get to a (possibly exising) ISA card need to disable
 * our current PCI card.
 */
/*
 * This is just for booting: we just want to catch pure
 * legacy vga therefore we don't worry about mmio etc.
 * This stuff should really go into vgaHW.c. However then
 * the driver would have to load the vga-module prior to
 * doing int10.
 */
/*void
LockLegacyVGA(int screenIndex,legacyVGAPtr vga)
{
    xf86SetCurrentAccess(FALSE, xf86Screens[screenIndex]);
    vga->save_msr = inb(0x3CC);
    vga->save_vse = inb(0x3C3);
    vga->save_46e8 = inb(0x46e8);
    vga->save_pos102 = inb(0x102);
    outb(0x3C2, ~(u8)0x03 & vga->save_msr);
    outb(0x3C3, ~(u8)0x01 & vga->save_vse);
    outb(0x46e8, ~(u8)0x08 & vga->save_46e8);
    outb(0x102, ~(u8)0x01 & vga->save_pos102);
    xf86SetCurrentAccess(TRUE, xf86Screens[screenIndex]);
}

void
UnlockLegacyVGA(int screenIndex, legacyVGAPtr vga)
{
    xf86SetCurrentAccess(FALSE, xf86Screens[screenIndex]);
    outb(0x102, vga->save_pos102);
    outb(0x46e8, vga->save_46e8);
    outb(0x3C3, vga->save_vse);
    outb(0x3C2, vga->save_msr);
    xf86SetCurrentAccess(TRUE, xf86Screens[screenIndex]);
}
*/

#if 0
int setup_int(struct int10_info * pInt)
{
    if (pInt != Int10Current) {
	if (!MapCurrentInt10(pInt))
	    return -1;
	Int10Current = pInt;
    }
    X86_EAX = (CARD32) pInt->ax;
    X86_EBX = (CARD32) pInt->bx;
    X86_ECX = (CARD32) pInt->cx;
    X86_EDX = (CARD32) pInt->dx;
    X86_ESI = (CARD32) pInt->si;
    X86_EDI = (CARD32) pInt->di;
    X86_EBP = (CARD32) pInt->bp;
    X86_ESP = 0x1000; X86_SS = pInt->stackseg >> 4;
    X86_EIP = 0x0600; X86_CS = 0x0;	/* address of 'hlt' */
    X86_DS = 0x40;			/* standard pc ds */
    X86_ES = pInt->es;
    X86_FS = 0;
    X86_GS = 0;
    X86_EFLAGS = X86_IF_MASK | X86_IOPL_MASK;

    return 0;
}

void finish_int(struct int10_info * pInt)
{
    pInt->ax = (CARD32) X86_EAX;
    pInt->bx = (CARD32) X86_EBX;
    pInt->cx = (CARD32) X86_ECX;
    pInt->dx = (CARD32) X86_EDX;
    pInt->si = (CARD32) X86_ESI;
    pInt->di = (CARD32) X86_EDI;
    pInt->es = (CARD16) X86_ES;
    pInt->bp = (CARD32) X86_EBP;
    pInt->flags = (CARD32) X86_FLAGS;
}
#endif
