#include <stdio.h>
#include <fcntl.h>
#include <types.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include "pci-userspace.h"

#define PCI_FILL_IDENT		1
#define PCI_FILL_IRQ		2
#define PCI_FILL_BASES		4
#define PCI_FILL_ROM_BASE	8
#define PCI_FILL_SIZES		16
#define PCI_FILL_CLASS		32
#define PCI_FILL_RESCAN		0x10000

#define DEBUG_PCI 1

struct pci_tag ltag;

int pciNumBuses = 0;
extern char *devpath;
extern int dev_pcidomain, dev_pcibus, dev_pcislot, dev_pcifn;

u16 swap16(u16 data)
{
  return ((data >> 8) & 0xff) | ((data << 8) & 0xff00);
}

u32 swap32(u32 data)
{
  return
    ((data >> 24) & 0x000000ff) |
    ((data >>  8) & 0x0000ff00) |
    ((data <<  8) & 0x00ff0000) |
    ((data << 24) & 0xff000000);
}

struct pci_tag *findPci(unsigned short bx)
{
	char buffer[200];
	struct pci_tag *tag = &ltag;

	int bus = (bx >> 8) & 0xFF;
	int slot = (bx >> 3) & 0x1F;
	int func = bx & 0x7;

	if (bus != dev_pcibus || slot != dev_pcislot)
	  return NULL;

	tag->bus = bus;
	tag->slot = slot;
	tag->func = func;

	snprintf(buffer, sizeof(buffer), "/sys/bus/pci/devices/%04x:%02x:%02x.%01x/config", dev_pcidomain, bus, slot, func);
	if (((tag->fdConfig = open(buffer, O_RDWR)) < 0))
		return NULL;

	read(tag->fdConfig, &tag->vendor, sizeof(tag->vendor));
	read(tag->fdConfig, &tag->device, sizeof(tag->device));

	return tag;
}

u32 pciSlotBX(struct pci_tag *tag)
{
	return (tag->bus << 8) | (tag->slot << 3) | (tag->func);
}

u8 pciReadByte(struct pci_tag *tag, u32 idx)
{
	u8 data;
	
	lseek(tag->fdConfig, idx, SEEK_SET);
	if (read(tag->fdConfig, &data, sizeof(data)) == sizeof(data))
		return data;
#ifdef DEBUG_PCI
	printf("PCI: device not found while read byte (%x:%x.%x)\n",
	       tag->bus, tag->slot, tag->func);
#endif
	return 0;
}

u16 pciReadWord(struct pci_tag *tag, u32 idx)
{
	u16 data;
	
	lseek(tag->fdConfig, idx, SEEK_SET);
	if (read(tag->fdConfig, &data, sizeof(data)) == sizeof(data))
		return swap16(data);
#ifdef DEBUG_PCI
	printf("PCI: device not found while read word (%x:%x.%x)\n",
	       tag->bus, tag->slot, tag->func);
#endif
	return 0;
}

u32 pciReadLong(struct pci_tag *tag, u32 idx)
{
	u32 data;
	
	lseek(tag->fdConfig, idx, SEEK_SET);
	if (read(tag->fdConfig, &data, sizeof(data)) == sizeof(data))
		return swap32(data);
#ifdef DEBUG_PCI
	printf("PCI: device not found while read long (%x:%x.%x)\n",
	       tag->bus, tag->slot, tag->func);
#endif
	return 0;
}


void pciWriteLong(struct pci_tag *tag, u32 idx, u32 data)
{
        data = swap32(data);
	lseek(tag->fdConfig, idx, SEEK_SET);
	if (write(tag->fdConfig, &data, sizeof(data)) != sizeof(data))
#ifdef DEBUG_PCI
		printf("PCI: device not found while write long (%x:%x.%x)\n",
		       tag->bus, tag->slot, tag->func);
#endif
}

void pciWriteWord(struct pci_tag *tag, u32 idx, u16 data)
{
        data = swap16(data);
	lseek(tag->fdConfig, idx, SEEK_SET);
	if (write(tag->fdConfig, &data, sizeof(data)) != sizeof(data))
#ifdef DEBUG_PCI
		printf("PCI: device not found while write word (%x:%x.%x)\n",
		       tag->bus, tag->slot, tag->func);
#endif

}

void pciWriteByte(struct pci_tag *tag, u32 idx, u8 data)
{
	lseek(tag->fdConfig, idx, SEEK_SET);
	if (write(tag->fdConfig, &data, sizeof(data)) != sizeof(data))
#ifdef DEBUG_PCI
		printf("PCI: device not found while write long (%x:%x.%x)\n",
		       tag->bus, tag->slot, tag->func);
#endif
}
