
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#if 0
#include <sys/io.h>
#endif
#include <sys/mman.h>
#include <sys/syscall.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <x86emu.h>
#include <dirent.h>
#include "post.h"
#include "pci-userspace.h"

unsigned char biosmem[1024 * 1024];
struct int10_info p;
struct int10_info * current = 0;
int verbose = 1;
void *iobase;
int dev_pcidomain, dev_pcibus, dev_pcislot, dev_pcifn;
void *vmem;
char *devpath;

/* Interrupt multiplexer */
static void do_int(int num)
{
	int ret = 0;

	printf("int%x vector at %x\n", num, getIntVect(num));

	/* This is a pInt leftover */
	current->num = num;

	switch (num) {
#ifndef _PC
	case 0x10:
	case 0x42:
	case 0x6D:

		if (getIntVect(num) == 0xFF065) {
			ret = int42_handler();
		}
		break;
#endif
	case 0x15:
		ret = int15_handler();
		break;
	case 0x16:
		ret = int16_handler();
		break;
	case 0x1A:
		ret = int1A_handler();
		break;
	case 0xe6:
		ret = intE6_handler();
		break;
	default:
		break;
	}

	if (!ret)
		ret = run_bios_int(num);

	if (!ret) {
		printf("\nint%x: not implemented\n", num);
		//x86emu_dump_xregs();
	}
}

static unsigned char *mapitin(char *file, off_t where, size_t size)
{
	void *z;

	int fd = open(file, O_RDWR, 0);
	if (fd < 0)
		die(file);
	
	z = mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, where);
	if (z == (void *) -1)
		die("mmap");
	
	close(fd);
	return z;

}

static char *getenvar(char *name)
{
	char *retval = getenv(name);
	if (retval)
		return retval;
	exit(-EINVAL);
}

static X86EMU_pioFuncs x_pio_funcs = {
	x_inb, x_inw, x_inl,
	x_outb, x_outw, x_outl
};

static X86EMU_memFuncs x_mem_funcs = {
	x_rdb, x_rdw, x_rdl,
	x_wrb, x_wrw, x_wrl
};

#if 0
#define NAME_SIZE 14
#define MAX_DEVICES 100
struct device {
	char name[NAME_SIZE];
	int parent;
	int level;
	int video;
};
int max_device = 0;
struct device devices[MAX_DEVICES];
#endif

static char buffer[4096];
#define PCI_ROOT "/sys/bus/pci/devices/"

#if 0
static void find_video(char *rootdir, int level, int parent)
{
	int i, fd;
s	DIR *root;
	struct dirent *dp;
	uint32_t class;
	
	root = opendir(rootdir);
	for (dp = readdir(root); dp != NULL; dp = readdir(root)) {
		if (strlen(dp->d_name) != 12)
			continue;
		if (dp->d_name[4] != ':')
			continue;
		
		for (i = 0; i < max_device; i++)
			if (strcmp(devices[i].name, dp->d_name) == 0)
				if (level > devices[i].level) {
					devices[i].parent = parent;
					devices[i].level = level;
					break;
				}
		if (i != max_device)
			continue;
		
		strcpy(devices[max_device].name, dp->d_name);
		devices[max_device].parent = parent;
		devices[max_device].level = level;
		
		strcpy(buffer, PCI_ROOT);
		strcat(buffer, dp->d_name);
		strcat(buffer, "/config");
		fd = open(buffer, O_RDONLY);
		lseek(fd, 8, SEEK_SET);
		read(fd, &class, sizeof(class));
		class = class >> 8;
		if (class == 0x30000)
			devices[max_device].video = 1;
		close(fd);
		
		max_device++;
		strcpy(buffer, rootdir);
		strcat(buffer, dp->d_name);
		strcat(buffer, "/");
		find_video(buffer, level + 1, max_device - 1);
	}
	closedir(root);
}
#endif

static void do_int10(void)
{
	X86_ESP = 0x1000; X86_SS = 0x30;/* pInt->stackseg >> 4; */
	X86_EIP = 0x0600; X86_CS = 0x0;	/* address of 'hlt' */
	X86_DS = 0x40;			/* standard pc ds */
	//	X86_EFLAGS = X86_IF_MASK | X86_IOPL_MASK;
	run_bios_int(0x10);
	X86EMU_exec();
}

extern void dprint(unsigned long start, unsigned long size);

static void check_vbe(void)
{
	u8 *buffer = &biosmem[0x10000];

	memset(buffer, 0, 512);
	buffer[0] = 'V';
	buffer[1] = 'B';
	buffer[2] = 'E';
	buffer[3] = '2';
	X86_EAX = 0x4f00;
	X86_EDI = 0x0;
	X86_ES = 0x1000;

	do_int10();

	printf("VESA test result: %x\n", X86_EAX);

	dprint((unsigned long)buffer, 512);
}

int main(int argc, char **argv)
{
	uint16_t bridge_ctl;
	int i, j;
	unsigned int entry, k;
	size_t size = 0;
	int base = 0;
	char *fsegname = 0;
	unsigned char *fsegptr;
	unsigned short initialip = 0, initialcs = 0, devfn = 0;
	X86EMU_intrFuncs intFuncs[256];
	int fd;
	unsigned char command, new_command;
	//	char *devpath = getenvar("DEVPATH");
	//	find_video(PCI_ROOT, 0, -1);
	
	if (argc > 1)
		devpath = argv[1];
	else
		devpath = getenvar("DEVPATH");

	strcpy(buffer, "");
	strncat(buffer, devpath, sizeof(buffer) - strlen(buffer) - 1);
	strncat(buffer, "/config", sizeof(buffer) - strlen(buffer) - 1);
	fd = open(buffer, O_RDWR);
	lseek(fd, 4, SEEK_SET);
	read(fd, &command, sizeof(command));
	lseek(fd, 4, SEEK_SET);
	new_command = 0x83;
	write(fd, &new_command, sizeof(command));
	close(fd);
	
	size = 32 * 1024;
	base = 0xc0000;
	initialcs = 0xc000;
	initialip = 0x0003;

	current = &p;
	memset(biosmem, 0xf4, sizeof(biosmem));
	X86EMU_setMemBase(biosmem, sizeof(biosmem));
	X86EMU_setupPioFuncs(&x_pio_funcs);
	X86EMU_setupMemFuncs(&x_mem_funcs);

#if 0
	ioperm(0, 0x400, 1);

	if (iopl(3) < 0) {
		warn("iopl failed, continuing anyway");
	}
#endif
	/* Emergency sync ;-) */
	sync();
	sync();

	/* Setting up interrupt environment.
	 * basically this means initializing PCI and
	 * intXX handlers.
	 */
	for (i = 0; i < 256; i++)
		intFuncs[i] = do_int;
	X86EMU_setupIntrFuncs(intFuncs);
	
	{
		char *p = devpath;
		unsigned int d,b,s,f;
		unsigned long iophys;

		for (d = 0; d < 4; d++) {
			while(*p != '/')
				p++;
			p++;
		}
		sscanf(p, "%x:%x:%x.%x", &d, &b, &s, &f);
		devfn = (b << 8) | (s << 3) | f;

		dev_pcibus = b;
		dev_pcislot = s;
		dev_pcifn = f;
		dev_pcidomain = d;
		iophys = syscall(200, 2, b, devfn & 0xff);
		printk("iophys = %08lx\n", iophys);
		iobase = mapitin("/dev/mem", iophys, 0x10000);
		
#if 0
		for (entry = 0; entry < max_device; entry++)
			if (strcmp(p, devices[entry].name) == 0)
				break;
#endif
	}
	current->ax = devfn;
	current->dx = 0x80;
	
	strcpy(buffer, "");
	strncat(buffer, devpath, sizeof(buffer) - strlen(buffer) - 1);
	strncat(buffer, "/rom", sizeof(buffer) - strlen(buffer) - 1);
	fd = open(buffer, O_RDWR);
	write(fd, "1\n", sizeof("1\n"));
	
	lseek(fd, 0, SEEK_SET);
	k = 0;
	while ((i = read(fd, buffer, sizeof(buffer))) > 0) {
		for (j = 0; j < i; j++)
			wrb(base + j + k, buffer[j]);
		k += sizeof(buffer);
	}
	close(fd);
	if ((i <= 0) && (k == 0)) {
		printf("Failed to read video ROM\n");
		exit(1);
	}
	
	strcpy(buffer, "");
	strncat(buffer, devpath, sizeof(buffer) - strlen(buffer) - 1);
	strncat(buffer, "/resource0", sizeof(buffer) - strlen(buffer) - 1);
	vmem = mapitin(buffer, 0, 0x200000);

#if 0
	/* temp code to make VGA routing work */
	for (i = 0; i < max_device; i++)
		if (devices[i].video) {
			strcpy(buffer, PCI_ROOT);
			strcat(buffer, devices[devices[i].parent].name);
			strcat(buffer, "/config");
			fd = open(buffer, O_RDWR);
			lseek(fd, 0x3e, SEEK_SET);
			read(fd, &bridge_ctl, sizeof(bridge_ctl));
			bridge_ctl = bridge_ctl & ~0x08;
			lseek(fd, 0x3e, SEEK_SET);
			write(fd, &bridge_ctl, sizeof(bridge_ctl));
			close(fd);
		}
#endif

#if 0
	strcpy(buffer, PCI_ROOT);
	strcat(buffer, devices[devices[entry].parent].name);
	strcat(buffer, "/config");
	fd = open(buffer, O_RDWR);
	lseek(fd, 0x3e, SEEK_SET);
	read(fd, &bridge_ctl, sizeof(bridge_ctl));
	bridge_ctl = bridge_ctl | 0x08;
	lseek(fd, 0x3e, SEEK_SET);
	write(fd, &bridge_ctl, sizeof(bridge_ctl));
	close(fd);
#endif       
	if (fsegname) {
		fsegptr = mapitin(fsegname, (off_t) 0, 0x10000);
		for (i = 0; i < 0x10000; i++)
			wrb(0xf0000 + i, fsegptr[i]);
	} else {
		char *date = "06/11/99";
		for (i = i; date[i]; i++)
			wrb(0xffff5 + i, date[i]);
		/* set up eisa ident string */
		strcpy(biosmem + 0x0FFD9, "PCI_ISA");

		/* write system model id for IBM-AT */
		*((unsigned char *) (biosmem + 0x0FFFE)) = 0xfc;
	}

	/* cpu setup */
	X86_AX = devfn ? devfn : 0xff;
	X86_DX = 0x80;
	X86_EIP = initialip;
	X86_CS = initialcs;

	/* Initialize stack and data segment */
	X86_SS = 0x0030;
	X86_DS = 0x0040;
	X86_SP = 0xfffe;
	/* We need a sane way to return from bios
	 * execution. A hlt instruction and a pointer
	 * to it, both kept on the stack, will do.
	 */
	pushw(0xf4f4);		/* hlt; hlt */
	pushw(X86_SS);
	pushw(X86_SP + 2);

	X86_ES = 0x0000;

//	M.x86.debug = DEBUG_DECODE_F;
	X86EMU_exec();

	printf("done !\n");

	check_vbe();

	return 0;
}
